# SPDX-FileCopyrightText: 2022-present Stéphane Bidoul <stephane.bidoul@acsone.eu>
# SPDX-FileCopyrightText: 2022-present ACSONE <https://acsone.eu>
#
# SPDX-License-Identifier: MIT

from pathlib import Path

import pytest


@pytest.fixture(scope="session")
def data_path() -> Path:
    return Path(__file__).parent.joinpath("data")



# SPDX-FileCopyrightText: 2022-present Stéphane Bidoul <stephane.bidoul@acsone.eu>
#
# SPDX-License-Identifier: MIT

import sys
from pathlib import Path
from typing import Iterator

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib

from manifestoo_core.addon import is_addon_dir


def get_project_name(root: str) -> str:
    pyproject_toml = Path(root) / "pyproject.toml"
    pyproject = tomllib.loads(pyproject_toml.read_text(encoding="utf-8"))
    return pyproject["project"]["name"]


def load_hatch_odoo_config(root: str) -> dict:
    pyproject_toml = Path(root) / "pyproject.toml"
    pyproject = tomllib.loads(pyproject_toml.read_text(encoding="utf-8"))
    return pyproject.get("tool", {}).get("hatch-odoo", {})


def iter_addons_dirs(root: str, config: dict) -> Iterator[Path]:
    addons_dirs = config.get("addons_dirs")
    if not addons_dirs:
        raise RuntimeError("missing tools.hatch-odoo.addons_dir in pyproject.toml")
    for addons_dir in [Path(root) / d for d in addons_dirs]:
        if not addons_dir.is_dir():
            continue
        yield addons_dir


def iter_addon_dirs(
    root: str,
    config: dict,
    allow_not_installable: bool,
) -> Iterator[Path]:
    for addons_dir in iter_addons_dirs(root, config):
        for addon_dir in addons_dir.iterdir():
            if is_addon_dir(addon_dir, allow_not_installable=allow_not_installable):
                yield addon_dir



# SPDX-FileCopyrightText: 2022-present Stéphane Bidoul <stephane.bidoul@acsone.eu>
# SPDX-FileCopyrightText: 2022-present ACSONE <https://acsone.eu>
#
# SPDX-License-Identifier: MIT

from email import message_from_bytes
from pathlib import Path

import pytest
from build import ProjectBuilder


@pytest.mark.parametrize(
    "project_name",
    ["project1", "project2", "project3", "project4", "project5", "project6"],
)
def test_odoo_addons_dependencies(
    project_name: str, data_path: Path, tmp_path: Path
) -> None:
    expected_dependencies = {
        "odoo<15.1dev,>=15.0a",
        "odoo-addon-mis-builder<15.1dev,>=15.0dev",
        "wrapt",
        "click-odoo-contrib",
    }
    builder = ProjectBuilder(data_path / project_name)
    metadata_dir = builder.metadata_path(tmp_path)
    metadata_path = Path(metadata_dir) / "METADATA"
    assert metadata_path.is_file()
    metadata = message_from_bytes(metadata_path.read_bytes())
    dependencies = set(metadata.get_all("Requires-Dist"))
    assert dependencies == expected_dependencies



# SPDX-FileCopyrightText: 2022-present Stéphane Bidoul <stephane.bidoul@acsone.eu>
# SPDX-FileCopyrightText: 2022-present ACSONE <https://acsone.eu>
#
# SPDX-License-Identifier: MIT

from typing import List

from hatchling.metadata.plugin.interface import MetadataHookInterface
from manifestoo_core.metadata import POST_VERSION_STRATEGY_NONE, metadata_from_addon_dir

from .config import iter_addon_dirs, load_hatch_odoo_config


class OdooAddonsDependenciesMetadataHook(MetadataHookInterface):
    """Hatch metadata hook to populate 'project.depencencies' from Odoo addons
    manifests."""

    PLUGIN_NAME = "odoo-addons-dependencies"

    def _get_odooo_addons_dependencies(self) -> List[str]:
        hatch_odoo_config = load_hatch_odoo_config(self.root)
        dependencies = set(hatch_odoo_config.get("dependencies", []))
        addon_dirs = list(
            iter_addon_dirs(
                self.root,
                hatch_odoo_config,
                # We don't want dependencies of addons that are not installable.
                allow_not_installable=False,
            )
        )
        depends_override = hatch_odoo_config.get("depends_override", {})
        for addon_dir in addon_dirs:
            addon_name = addon_dir.name
            # Do not add dependencies on addons that are in the project.
            depends_override[addon_name] = None
        options = dict(
            hatch_odoo_config,
            depends_override=depends_override,
            post_version_strategy_override=POST_VERSION_STRATEGY_NONE,
        )
        for addon_dir in addon_dirs:
            try:
                addon_metadata = metadata_from_addon_dir(addon_dir, options)
            except Exception as e:
                raise RuntimeError(
                    f"Error while generating metadata from {addon_dir}"
                ) from e
            addon_dependencies = addon_metadata.get_all("Requires-Dist", [])
            dependencies.update(addon_dependencies)
        return list(dependencies)

    def update(self, metadata: dict) -> None:
        """Update the project table's metadata."""
        if "dependencies" in metadata:
            raise ValueError(
                "'dependencies' may not be listed in the 'project' table when using "
                "hatch-odoo to populate dependencies from Odoo addons manifests. "
                "If you need to add dependencies that are not in Odoo addons "
                "manifests, please use the 'tools.hatch-odoo.dependencies' key."
            )
        if "dependencies" not in metadata.get("dynamic", []):
            raise ValueError(
                "'dependencies' must be listed in 'project.dynamic' when using "
                "hatch-odoo to populate dependencies from Odoo addons manifests."
            )
        metadata["dependencies"] = self._get_odooo_addons_dependencies()



# SPDX-FileCopyrightText: 2022-present Stéphane Bidoul <stephane.bidoul@acsone.eu>
# SPDX-FileCopyrightText: 2022-present ACSONE <https://acsone.eu>
#
# SPDX-License-Identifier: MIT

import subprocess
import sys
import zipfile
from pathlib import Path

import pytest


@pytest.mark.parametrize(
    "project, addons_only",
    [
        ("project1", False),
        ("project2", False),
        ("project3", False),
        ("project4", False),
        ("project5", False),
        ("project6", True),
    ],
)
@pytest.mark.parametrize("build_via_sdist", [True, False])
def test_build(
    project: str,
    addons_only: bool,
    build_via_sdist: bool,
    data_path: Path,
    tmp_path: Path,
) -> None:
    build_cmd = [
        sys.executable,
        "-m",
        "build",
        "-n",
        "-o",
        str(tmp_path),  # str for compat with Python 3.7 on Windows
    ]
    if not build_via_sdist:
        build_cmd.append("-w")
    subprocess.run(
        build_cmd,
        check=True,
        cwd=data_path / project,
    )
    wheel_file = next(tmp_path.glob(f"{project}-*.whl"))
    with zipfile.ZipFile(wheel_file) as zip_file:
        files = set(zip_file.namelist())
        for addon_name in ("addona", "addonb", "addon_uninstallable"):
            assert f"odoo/addons/{addon_name}/__init__.py" in files
            assert f"odoo/addons/{addon_name}/__manifest__.py" in files
        assert addons_only or f"{project}/__init__.py" in files



# SPDX-FileCopyrightText: 2022-present Stéphane Bidoul <stephane.bidoul@acsone.eu>
# SPDX-FileCopyrightText: 2022-present ACSONE <https://acsone.eu>
#
# SPDX-License-Identifier: MIT

import subprocess
import sys
from pathlib import Path
from typing import List

import pytest


@pytest.mark.parametrize(
    "project_name,expected_editable_pth_lines",
    [
        ("project1", ["src"]),
        ("project2", ["src", "build/__editable_odoo_addons__"]),
        ("project3", [""]),
        ("project4", ["src", "addons_group1", "addons_group2"]),
        ("project5", ["", "build/__editable_odoo_addons__"]),
        ("project6", ["build/__editable_odoo_addons__"]),
    ],
)
def test_odoo_addons_dependencies(
    project_name: str,
    expected_editable_pth_lines: List[str],
    data_path: Path,
    tmp_path: Path,
) -> None:
    subprocess.run(
        [
            sys.executable,
            "-m",
            "pip",
            "install",
            "-e",
            str(data_path / project_name),  # str for compat with Python 3.7 on Windows
            "--no-deps",
            "--no-build-isolation",
            "--target",
            str(tmp_path),  # str for compat with Python 3.7 on Windows
        ],
        check=True,
    )
    # Check we have the .pth lines we expect.
    pth_lines = set()
    for pth_file in tmp_path.glob("*.pth"):
        pth_lines.update((tmp_path / pth_file).read_text().splitlines())
    assert pth_lines == {
        str(data_path / project_name / line) for line in expected_editable_pth_lines
    }
    # Check all addons are in the editable paths.
    expected_editable_addon_names = ["addona", "addonb"]
    editable_addon_names = []
    for pth_line in pth_lines:
        addons_dir = Path(pth_line) / "odoo" / "addons"
        for addon_name in expected_editable_addon_names:
            if (addons_dir / addon_name).is_dir():
                editable_addon_names.append(addon_name)
    assert sorted(editable_addon_names) == expected_editable_addon_names
    # Check that the project odoo packages have not been copied to the target.
    assert not tmp_path.joinpath("odoo").exists()
    assert not tmp_path.joinpath(project_name).exists()



# SPDX-FileCopyrightText: 2022-present Stéphane Bidoul <stephane.bidoul@acsone.eu>
# SPDX-FileCopyrightText: 2022-present ACSONE <https://acsone.eu>
#
# SPDX-License-Identifier: MIT

import atexit
import shutil
import tempfile
from pathlib import Path
from typing import Any

from hatchling.builders.hooks.plugin.interface import BuildHookInterface

from .config import get_project_name, iter_addon_dirs, load_hatch_odoo_config


class OdooAddonsDirsBuildHook(BuildHookInterface):
    """Hatch build hook to add Odoo addons directories to the wheel, and available as
    symlinks in sys.path in editable mode."""

    PLUGIN_NAME = "odoo-addons-dirs"

    def initialize(self, version: str, build_data: dict[str, Any]) -> None:
        if self.target_name != "wheel":
            return
        hatch_odoo_config = load_hatch_odoo_config(self.root)
        if version == "standard":
            force_include = build_data["force_include"]
            for addon_dir in iter_addon_dirs(
                self.root,
                hatch_odoo_config,
                # We force-include addons that are installable False to avoid that the
                # default hatch behaviour adds them at the wrong place in the wheel.
                allow_not_installable=True,
            ):
                addon_name = addon_dir.name
                force_include[addon_dir] = f"odoo/addons/{addon_name}"
        elif version == "editable" and self.config.get("editable_symlinks", True):
            editable_path = Path(self.root) / "build" / "__editable_odoo_addons__"
            if editable_path.is_dir():
                shutil.rmtree(editable_path)
            editable_odoo_addons_path = editable_path / "odoo" / "addons"
            has_editable_symlinks = False
            for addon_dir in iter_addon_dirs(
                self.root,
                hatch_odoo_config,
                allow_not_installable=False,
            ):
                # Symlink the addon in the editable path.
                if not has_editable_symlinks:
                    editable_odoo_addons_path.mkdir(parents=True)
                    has_editable_symlinks = True
                addon_name = addon_dir.name
                editable_addon_path = editable_odoo_addons_path / addon_name
                editable_addon_path.symlink_to(addon_dir)
            # Add .pth to build/__editable_odoo_addons__ in wheel.
            if has_editable_symlinks:
                with tempfile.NamedTemporaryFile(
                    mode="w", encoding="utf-8", delete=False
                ) as pth_file:
                    atexit.register(Path(pth_file.name).unlink)
                    pth_file.write(str(editable_path) + "\n")
                project_name = get_project_name(self.root)
                force_include_editable = build_data["force_include_editable"]
                force_include_editable[pth_file.name] = (
                    f"{project_name}_editable_odoo_addons.pth"
                )
                # work around for https://github.com/acsone/hatch-odoo/issues/12
                force_include = build_data["force_include"]
                force_include[pth_file.name] = (
                    f"{project_name}_editable_odoo_addons.pth"
                )



# SPDX-FileCopyrightText: 2022-present Stéphane Bidoul <stephane.bidoul@acsone.eu>
# SPDX-FileCopyrightText: 2022-present ACSONE <https://acsone.eu>
#
# SPDX-License-Identifier: MIT

from typing import Type

from hatchling.plugin import hookimpl

from .build_hook import OdooAddonsDirsBuildHook
from .metadata_hook import OdooAddonsDependenciesMetadataHook


@hookimpl
def hatch_register_build_hook() -> Type[OdooAddonsDirsBuildHook]:
    return OdooAddonsDirsBuildHook


@hookimpl
def hatch_register_metadata_hook() -> Type[OdooAddonsDependenciesMetadataHook]:
    return OdooAddonsDependenciesMetadataHook
